# -*- coding: utf-8 -*-
'''*****************************************************************************
***  This file is part of HgWeb Manager.                                     ***
***                                                                          ***
***  Copyright (C) 2014 - 2015                                               ***
***  CodeGNU Solutions <Licensing _AT_ CodeGNU _DOT_ com>                    ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*****************************************************************************'''

from manrepolock import RepositoryManagerLock
import config


settings = config.Manager()


def plugInEnable(repository, plugInName, scriptFile):
	""" Enables a plug-in to a managed repository.

	@pre The repository must already exist.

	@param[in] repository		The targeted repository to add a plug-in to.
	@param[in] plugInName		The name of the plug-in to add.
	@param[in] scriptFile		The file path location of the plug-in script.

	@post The managed repository has it's configuration file changed which
	enables the supplied plug-in.

	@throws ValueError		When the repository storage name does not exist.
	"""
	from manrepo import RepositoryPlugin

	with RepositoryManagerLock():
		print "\tEnable plug-in %s for %s" % (plugInName, repository.StorageName)
		if not __doesRepositoryStorageNameExist(repository):
			raise ValueError("The repository '%s' does not exist." % repository.StorageName)

		print 'Creating plug-in information ...'
		plugIn = RepositoryPlugin(plugInName, scriptFile)
		repository.addPlugIn(plugIn)
		__writeRepositoryDetails(repository)


def plugInDisable(repository, named):
	""" Disables a plug-in to a managed repository.

	@pre The repository must already exist.

	@param[in] repository		The targeted repository to remove a plug-in from.
	@param[in] named			The name of the plug-in to remove.

	@post The managed repository has it's configuration file changed which
	disables the supplied plug-in.

	@throws ValueError		When the repository storage name does not exist.
	"""
	with RepositoryManagerLock():
		print "\tDisable plug-in %s for %s" % (named, repository.StorageName)
		if not __doesRepositoryStorageNameExist(repository):
			raise ValueError("The repository '%s' does not exist." % repository.StorageName)

		print 'Removing plug-in information ...'
		repository.removePlugIn(named)
		__writeRepositoryDetails(repository)


def setPlugInSetting(repository, plugIn, named, value):
	with RepositoryManagerLock():
		print "\t Setting repository %s's '%s' plug-in setting '%s' to %s" % (repository.StorageName, plugIn, named, value)
		if not __doesRepositoryStorageNameExist(repository):
			raise ValueError("The repository '%s' does not exist." % repository.StorageName)

		print 'Modifying plug-in setting ...'
		repository.setSetting(plugIn, named, value)
		__writeRepositoryDetails(repository)


def create(repository, ignoreGroups, skipCommon):
	""" Creates a new repository under HgWeb.

	@pre The repository to create cannot already exist.

	@param[in] repository		The targeted repository to create.
	@param[in] ignoreGroups		The list of all ignore groups to add.
	@param[in] skipCommon		A flag that when true indicates the common group
								should not be added by default.

	@post The file system now contains a new directory in the repository path
	that is the newly generated repository. The new repository is registered
	in HgWeb to display and is secured to prevent accidental modification.

	@throws ValueError		When the repository storage or display name exists.
	@throws RuntimeError	When the new repository creation encounters an
							error.
	"""
	import os
	import shutil
	import stat

	with RepositoryManagerLock():
		print "\tAdding repository: %s" % repository.StorageName
		if __doesRepositoryStorageNameExist(repository):
			raise ValueError("The repository storage name '%s' already exists." % repository.StorageName)

		if __doesRepositoryDisplayNameExist(repository):
			raise ValueError("The display name '%s' already exists." % repository.DisplayName)

		try:
			__addRepository(repository)
			__create_ignores(repository, ignoreGroups, skipCommon)
		except RuntimeError:
			__removeRepository(repository)
			pass

		try:
			__writeRepositoryDetails(repository)
			__registerRepository(repository)
		except RuntimeError:
			os.chmod(os.path.join(settings.RepositoryPath, repository.StorageName, '.hg', "hgrc"), stat.S_IWRITE)
			shutil.rmtree(settings.RepositoryPath + os.sep + repository.StorageName)
			pass


def register(repository):
	""" Connects an existing repository to the mercurial web registry.

	@pre The repository to register must already exist.

	@param[in] repository	The targeted repository to register.

	@post The new repository is registered in HgWeb to display and is
	secured to prevent accidental modification.

	@throws ValueError		When the repository storage or display name does not
							exist.
	@throws RuntimeError	When the new repository registration encounters an
							error.
	"""
	with RepositoryManagerLock():
		print "\tRegistering the repository: %s" % repository.StorageName
		if __doesRepositoryStorageNameExist(repository):
			__registerRepository(repository)
		else:
			raise ValueError("The repository '%s' does not exists." % repository.StorageName)


def modify(newRepository, currentStorageName, ignoreGroups, skipCommon):
	""" Changes an existing repository in HgWeb.

	@pre The repository to modify must already exist.

	@param[in] newRepository		The new values for the repository to
									modify.
	@param[in] currentStorageName	The current storage name of the
									repository to modify.
	@param[in] ignoreGroups			The list of all ignore groups to add.
	@param[in] skipCommon			A flag that when true indicates the
									common group should not be added by
									default.

	@post The repository is changed with the requested modifications.

	@throws ValueError		When the repository storage name does not exist.
	@throws RuntimeError	When the new repository modification encounters an
							error.
	"""
	from manrepo import Repository

	with RepositoryManagerLock():
		print "\tModifying repository: %s" % currentStorageName
		oldRepository = Repository(currentStorageName)

		if not __doesRepositoryStorageNameExist(oldRepository):
			raise ValueError("The repository '%s' was not found." % oldRepository.StorageName)

		__unregisterRepository(oldRepository)

		if __doesRepositoryDisplayNameExist(newRepository):
			__registerRepository(oldRepository)
			raise ValueError("The display name '%s' already exists." % newRepository.DisplayName)

		# Update renamed repositories
		if not oldRepository == newRepository:
			try:
				__renameStorage(oldRepository, newRepository.StorageName)
			except RuntimeError:
				__registerRepository(oldRepository)
				pass

		# Fill out the details of about the new repository.
		try:
			__writeRepositoryDetails(newRepository)
			# Update ignore file if requested.
			if ignoreGroups is not None:
				__create_ignores(newRepository, ignoreGroups, skipCommon)
		except RuntimeError:
			__registerRepository(oldRepository)
			print "Successfully rolled backed action(s)."
			pass

		# Add the new repository to the web registry.
		__registerRepository(newRepository)


def delete(repository):
	""" Removes an existing repository.

	@pre The repository to remove must already exist.

	@param[in] repository	The targeted repository to delete.

	@post The file system no longer contains a directory in the repository
	path that is the supplied repository. The repository is no longer
	registered in HgWeb.

	@throws ValueError		When the repository storage name does not exist.
	"""
	with RepositoryManagerLock():
		if not __doesRepositoryStorageNameExist(repository):
			raise ValueError("The repository %s was not found." % repository.StorageName)

		print "\tRemoving repository: %s" % repository.StorageName
		__unregisterRepository(repository)
		__removeRepository(repository)


def list():
	""" Outputs a listing of all the known managed repositories """
	import manrepo

	print "Managed Repositories:"
	repoCol = manrepo.ManagedCollection()
	for repo in repoCol.Repositories:
		print "\t%s (%s)" % (repo.StorageName, repo.DisplayName)


#               Functions For Fulfilling Repository Management               ###


def __create_ignores(repo, ignoreGroups, skipCommon):
	""" Generates and commits an ignore file to an existing repository.

	@pre The repository to create an ignore file must already exist. The list
	of ignore patterns to use is expected to be in an XML file named
	"ignores.xml" and be in the hwm-ignore schema.

	@param[in] repo				The targeted repository for the ignore file.
	@param[in] ignoreGroups		The list of all ignore groups to add.
	@param[in] skipCommon		A flag that when true indicates the common group
								should not be added by default.

	@post The managed repository now has an ignore file committed by the
	HgWeb manager.

	@throws RuntimeError	When the list of ignores does not have a group.
	"""
	import os
	import ignorepo

	# Extract Ignore Collection
	if ignoreGroups is None:
		ignoreGroups = []
	ignores = ignorepo.extract(os.path.join(settings.DataPath, 'ignores.xml'))

	if skipCommon:
		del ignores['Common']
	elif 'common' not in [ignoreGroup.lower() for ignoreGroup in ignoreGroups]:
		ignoreGroups.append('Common')

	if '+' not in ignoreGroups:
		for ignoreGroup in ignoreGroups:
			if not ignoreGroup.lower() in [ignore.lower() for ignore in ignores]:
				raise RuntimeError("The list of ignores does not have the '%s' group." % ignoreGroup)

	ignoreFile = settings.RepositoryPath + os.sep + repo.StorageName + os.sep + '.hgignore'
	wasCreated = not os.path.isfile(ignoreFile)
	if wasCreated:
		print "Creating ignore file ..."
	else:
		print "Rewriting ignore file ..."

	# Write Ignore File
	with open(ignoreFile, 'w') as hgIgnore:
		for groupName, groupPatterns in ignores.items():
			if '+' not in ignoreGroups and not groupName.lower() in [ignoreGroup.lower() for ignoreGroup in ignoreGroups]:
				continue
			hgIgnore.write("#\n#\tIgnore Group: %s\n#" % groupName)
			# Reverse the sorting order to put the regular expressions first.
			groupPatterns.sort(key=lambda p: p.Filter, reverse=True)
			filteringOn = None
			for pattern in groupPatterns:
				if pattern.Filter == filteringOn:
					hgIgnore.write("\n%s\t\t#  %s" % (pattern.Value, pattern.Description))
				elif filteringOn is None:
					hgIgnore.write("\nsyntax: %s\n%s\t\t#  %s" % (ignorepo.ignore_syntax_name(pattern.Filter), pattern.Value, pattern.Description))
				else:
					hgIgnore.write("\n\nsyntax: %s\n%s\t\t#  %s" % (ignorepo.ignore_syntax_name(pattern.Filter), pattern.Value, pattern.Description))
				filteringOn = pattern.Filter
			hgIgnore.write('\n\n')

	if wasCreated:
		print "Adding generated ignores ..."
		__addFile(repo, '.hgignore')

	print "Committing generated ignores ..."
	__commitChanges(repo, 'Created generated ignores filter.')


def __commitChanges(repo, message):
	""" Commits all changes in a managed Mercurial repository as the HgWeb
	manager.

	@pre The requested repository must already exist.

	@param[in] repo		The targeted repository for which to commit.
	@param[in] message	The text to use as a commit message.

	@post Upon any error the standard error buffer contains an error message.

	@throws RuntimeError	When the added files fail to commit.
	"""
	import subprocess
	import os

	statusCode = subprocess.call([settings.HgCommand, '--repository', settings.RepositoryPath + os.sep + repo.StorageName, 'commit', '--message', message, '--user', settings.HgUser + ' - HgWeb Manager'])

	# Zero being a successful commit and one being nothing changed.
	if not ((0 == statusCode) or (1 == statusCode)):
		raise RuntimeError("Failed to commit files. Status code: " + statusCode)


def __addFile(repo, fileName):
	""" Marks a file as added in a managed Mercurial repository.

	@pre The requested repository must already exist.

	@param[in] repo			The targeted repository for which to add the file.
	@param[in] fileName		The directory path and the file name to add
							relative to the base of the repository.

	@post Upon any error the standard error buffer contains an error message.

	@throws RuntimeError	When the the news files fail to be added.
	"""
	import subprocess
	import os

	statusCode = subprocess.call([settings.HgCommand, '--repository', settings.RepositoryPath + os.sep + repo.StorageName, 'add', settings.RepositoryPath + os.sep + repo.StorageName + os.sep + fileName])

	# Zero being a successful commit and one being nothing to add.
	if not ((0 == statusCode) or (1 == statusCode)):
		raise RuntimeError("Failed to commit files. Status code: " + statusCode)


def __renameStorage(oldRepo, newName):
	""" Changes the storage name of an existing repository.

	@param[in] oldRepo		The existing repository to whose storage name is
							expected to be changed.
	@param[in] newName		The new storage name for the repository.

	@throws ValueError	When the supplied new storage name already exists.
	"""
	import os
	import shutil

	if not oldRepo.StorageName == newName:
		print "Renaming repository %s to %s ..." % (oldRepo.StorageName, newName)
		if os.path.isdir(settings.RepositoryPath + os.sep + newName):
			raise ValueError("Cannot rename the existing repository '%s' because the name '%s' already exists." % (oldRepo.StorageName, newName))
		shutil.move(settings.RepositoryPath + os.sep + oldRepo.StorageName, settings.RepositoryPath + os.sep + newName)


def __doesRepositoryStorageNameExist(repo):
	""" Checks to see if a repository storage name exists.

	@param[in] repo		The targeted repository for which to check.

	@return Gives true when it exists, false else-wise.
	"""
	import os

	if (repo is None) or (repo.StorageName is None):
		return False
	elif os.path.isdir(settings.RepositoryPath + os.sep + repo.StorageName + os.sep + '.hg'):
		return True
	else:
		return False


def __doesRepositoryDisplayNameExist(repo):
	""" Checks to see if a repository display name exists.

	@param[in] repo		The targeted repository for which to check.

	@return Gives true when it exists, false else-wise.
	"""
	from manrepo import ManagedCollection

	if (repo is None) or (repo.StorageName is None) or (repo.DisplayName is None):
		return False

	repoCol = ManagedCollection()
	for testRepo in repoCol.Repositories:
		if (testRepo.DisplayName is not None) and (not testRepo == repo) and (testRepo.DisplayName.lower() == repo.DisplayName.lower()):
			return True
	return False


def __unregisterRepository(repo):
	""" Removes the registration of the supplied repository.

	@param[in] repo		The targeted repository for which to unregister.

	@post The HgWeb configuration file does not contain the repository details
	and is returned to a read-only state. Upon any error the standard error
	buffer contains an error message.

	@throws RuntimeError	When the web path is missing the configuration file.
	@throws RuntimeError	When the configuration file does not contain a path
							section.
	"""
	from ConfigParser import SafeConfigParser
	import os

	parser = SafeConfigParser()
	if not parser.read(settings.HgWebPath + os.sep + 'hgweb.config'):
		raise RuntimeError("Failed to locate HgWeb configuration file.")

	if parser.has_section('paths'):
		if parser.has_option('paths', repo.StorageName):
			print "Unregistering repository ..."
			parser.remove_option('paths', repo.StorageName)
			with open(settings.HgWebPath + os.sep + 'hgweb.config', 'wb') as configfile:
				parser.write(configfile)
		else:
			print "The repository already is unregistered."
	else:
		raise RuntimeError("The HgWeb configuration file does not have a paths section! Failed to unregister repository.")


def __removeRepository(repo):
	""" Removes an existing Mercurial repository.

	@pre The requested repository to delete must already exist.

	@param[in] repo		The targeted repository for which to delete.

	@post Upon any error the standard error buffer contains an error message.
	"""
	import shutil
	import os
	import stat

	print "Removing repository directories ..."
	try:
		os.chmod(settings.RepositoryPath + os.sep + repo.StorageName + os.sep + '.hg' + os.sep + "hgrc", stat.S_IWRITE)
	except OSError as e:
		# If it's anything but the "No such file or directory", notify the user.
		if not 2 == e.errno:
			pass
	shutil.rmtree(settings.RepositoryPath + os.sep + repo.StorageName)


def __addRepository(repo):
	""" Initialises a new Mercurial repository.

	@pre The requested repository to create cannot already exist.

	@param[in] repo	The targeted repository for which to create.

	@post Upon any error the standard error buffer contains an error message.

	@throws ValueError		When the storage name is missing.
	@throws RuntimeError	When the initialisation call fails.
	"""
	import subprocess
	import os

	print "Initialising repository directory ..."
	if repo.StorageName is None:
		raise ValueError('Cannot create a managed repository without a storage name.')
	statusCode = subprocess.call([settings.HgCommand, 'init', settings.RepositoryPath + os.sep + repo.StorageName])
	if not (0 == statusCode):
		raise RuntimeError("Failed to initialise repository. Status code: " + statusCode)


def __writeRepositoryDetails(repo):
	""" Writes the details of a repository to the repository's
	configuration file.

	@param[in] repo		The targeted repository for which to write the
						configuration details.

	@post The HgWeb configuration file contains the new repository details and
	is returned to a read-only state.
	"""
	import os
	import stat
	import sys

	print "Saving repository configuration details ..."
	config = settings.RepositoryPath + os.sep + repo.StorageName + os.sep + '.hg' + os.sep + 'hgrc'

	if os.path.exists(config):
		hgrcMode = os.stat(config)[stat.ST_MODE]
		os.chmod(config, hgrcMode | stat.S_IWRITE)
	else:
		hgrcMode = stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH
	repo.save()
	# Ensure the repository details isn't accidentally modified.
	os.chmod(config, hgrcMode & (~stat.S_IWUSR & ~stat.S_IWGRP & ~stat.S_IWOTH))

	# Ensure the repository isn't accidentally deleted, moved, or blocked in Windows.
	if "win32" == sys.platform:
		__guardWindowsRepository(repo)
		print 'Secured repository store in Windows.'


def __guardWindowsRepository(repo):
	""" Guards the supplied repository for a repository on a Windows system.

	@param[in] repo	The targeted repository for which to secure.
	"""
	import subprocess
	import os

	print 'Securing for Windows ...'
	status = subprocess.call(["attrib.exe", '+H', '+I', '+S', settings.RepositoryPath + os.sep + repo.StorageName + os.sep + '.hg'])

	if not (0 == status):
		raise RuntimeError("Failed to secure repository store for windows. Status code: " + status)


def __registerRepository(repo):
	""" Adds the registration details of the repository.

	@param[in] repo		The targeted repository for which to register in HgWeb.

	@post The HgWeb configuration file contains the new repository details and
	is returned to a read-only state. Upon any error the standard error
	buffer contains an error message.

	@throws RuntimeError	When the HgWeb configuration file cannot be found.
	@throws ValueError		When the supplied repository is already registered.
	"""
	from ConfigParser import SafeConfigParser
	import os

	parser = SafeConfigParser()
	if not parser.read(settings.HgWebPath + os.sep + 'hgweb.config'):
		raise RuntimeError("Failed to locate HgWeb configuration file.")

	if parser.has_section('paths'):
		if parser.has_option('paths', repo.StorageName):
			raise ValueError("The repository already is registered.")
	else:
		parser.add_section('paths')
	print "Registering repository ..."
	parser.set('paths', repo.StorageName, settings.RepositoryPath + os.sep + repo.StorageName)
	with open(settings.HgWebPath + os.sep + 'hgweb.config', 'wb') as configfile:
		parser.write(configfile)
